""" Module to select best snapshot "send" commands to use for synchronization.

Based on optimzing a Directed Acyclic Graph (DAG),
where snapshots are the nodes,
and "send" diffs are the directed edges.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
"""

# import pprint
import Store

import logging
logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')


class _Node:

    def __init__(self, uuid, intermediate=False):
        self.uuid = uuid
        self.intermediate = intermediate
        self.previous = None
        self.diffSink = None
        self.diffSize = None

    def __unicode__(self):
        try:
            ancestors = ", %d ancestors" % (self.height-1)
        except AttributeError:
            ancestors = ""

        return u"%s from %s (%s%s) in %s" % (
            self._getPath(self.uuid),
            self._getPath(self.previous),
            Store.humanize(self.diffSize),
            ancestors,
            self.diffSink
        )

    def __str__(self):
        return unicode(self).encode('utf-8')

    @staticmethod
    def getPath(node):
        """ Return printable description of node, if not None. """
        if node is None:
            return None
        uuid = node.uuid
        return node._getPath(uuid)

    def _getPath(self, uuid):
        """ Return printable description of uuid. """
        result = Store.printUUID(uuid)
        try:
            result = "%s (%s)" % (self.diffSink.getVolume(uuid)['path'], result)
        except (KeyError, AttributeError):
            pass
        return result

    @staticmethod
    def summary(nodes):
        count = 0
        size = 0
        sinks = {}
        for n in nodes:
            count += 1

            if n.diffSize is not None:
                size += n.diffSize
                if n.diffSink in sinks:
                    sinks[n.diffSink] += n.diffSize
                else:
                    sinks[n.diffSink] = n.diffSize

        return {"count": count, "size": size, "sinks": sinks}


class BestDiffs:

    """ This analyzes and stores an optimal network (tree).

    The nodes are the desired (or intermediate) volumes.
    The directed edges are diffs from an available sink.
    """

    def __init__(self, volumes, delete=False):
        """ Initialize.

        volumes are the required snapshots.
        """
        self.nodes = {volume: _Node(volume, False) for volume in volumes}
        self.dest = None
        self.delete = delete

    def analyze(self, *sinks):
        """  Figure out the best diffs to use to reach all our required volumes. """
        # Use destination (already uploaded) edges first
        sinks = list(sinks)
        sinks.reverse()
        self.dest = sinks[0]

        nodes = [None]
        height = 1

        def sortKey(node):
            if node is None:
                return None
            return (node.intermediate, self._totalSize(node))

        while len(nodes) > 0:
            logger.debug("Analyzing %d nodes for height %d...", len(nodes), height)

            nodes.sort(key=sortKey)

            for fromNode in nodes:
                if self._height(fromNode) >= height:
                    continue

                fromSize = self._totalSize(fromNode)
                fromUUID = fromNode.uuid if fromNode else None

                logger.debug(
                    "Examining edges from %s (total %s)",
                    _Node.getPath(fromNode), Store.humanize(fromSize)
                )

                for sink in sinks:
                    logger.debug(
                        "Examining edges in %s",
                        sink
                    )

                    for edge in sink.iterEdges(fromUUID):
                        toUUID = edge['to']

                        # Skip any edges already in the destination
                        if sink != self.dest and self.dest.hasEdge(toUUID, fromUUID):
                            continue

                        if toUUID in self.nodes:
                            toNode = self.nodes[toUUID]
                        else:
                            toNode = _Node(toUUID, True)
                            self.nodes[toUUID] = toNode

                        newCost = self._cost(sink, edge['size'], fromSize, height)
                        if toNode.diffSink is None:
                            oldCost = None
                        else:
                            oldCost = self._cost(toNode.diffSink, toNode.diffSize, fromSize, self._height(toNode))

                        # Don't use a more-expensive path
                        if oldCost is not None and oldCost <= newCost:
                            continue

                        # Don't create circular paths
                        if self._wouldLoop(fromUUID, toUUID):
                            logger.debug("Ignoring looping edge: %s", toNode._getPath(edge['to']))
                            continue

                        logger.debug(
                            "Replacing edge (%s -> %s cost) %s",
                            Store.humanize(oldCost),
                            Store.humanize(newCost),
                            toNode
                        )

                        toNode.previous = fromUUID
                        toNode.diffSink = sink
                        toNode.diffSize = edge['size']

            nodes = [node for node in self.nodes.values() if self._height(node) == height]
            height += 1

        self._prune()

        for node in self.nodes.values():
            node.height = self._height(node)

    def _height(self, node):
        if node is None:
            return 0
        else:
            return 1 + self._height(self._getNode(node.previous))

    def _totalSize(self, node):
        if node is None:
            return 0

        prevSize = self._totalSize(self._getNode(node.previous))

        return (node.diffSize or 0) + prevSize

    def _getNode(self, uuid):
        return self.nodes[uuid] if uuid is not None else None

    def _wouldLoop(self, fromNode, toNode):
        if toNode is None:
            return False

        while fromNode is not None:
            if fromNode == toNode:
                return True

            fromNode = self.nodes[fromNode].previous

        return False

    def iterDiffs(self):
        """ Return all diffs used in optimal network. """
        nodes = self.nodes.values()
        nodes.sort(key=lambda node: self._height(node))
        for node in nodes:
            yield node
            # yield { 'from': node.previous, 'to': node.uuid, 'sink': node.diffSink,

    def summary(self):
        """ Return summary count and size in a dictionary. """
        return _Node.summary(self.nodes.values())

    def _prune(self):
        """ Get rid of all intermediate nodes that aren't needed. """
        done = False
        while not done:
            done = True
            for node in [node for node in self.nodes.values() if node.intermediate]:
                if not [dep for dep in self.nodes.values() if dep.previous == node.uuid]:
                    logger.debug("Removing unnecessary node %s", _Node.getPath(node))
                    del self.nodes[node.uuid]
                    done = False

    def _cost(self, sink, size, prevSize, height):
        cost = 0

        # Transfer
        cost += size if sink != self.dest else 0

        # Storage
        cost += size if self.delete or sink != self.dest else 0

        # Corruption risk
        cost += (prevSize + size) * (2 ** (height - 8))

        return cost



""" Abstract and component classes for sources and sinks.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
"""

# sink, source, src, dest: store
# volume, diff

# Classes: CamelCase
# files: CamelCase.py
# project: buttersink

# Volume: { uuid, path, size }
# Diff: { toUUID, fromUUID, size }

from __future__ import division

import math


class Store:

    """ Abstract class for storage back-end. """

    def hasEdge(self, toUUID, fromUUID):
        """ Store already contains this edge. """
        raise NotImplementedError

    def receive(self, toUUID, fromUUID, path):
        """ Return a file-like (stream) object to store a diff. """
        raise NotImplementedError

    def send(self, toUUID, fromUUID, streamContext, progress=True):
        """ Write the diff (toUUID from fromUUID) to the stream context manager. """
        raise NotImplementedError


def printBytes(number):
    """ Return a human-readable string for number. """
    return humanize(number)


def humanize(number):
    """ Return a human-readable string for number. """
    # units = ('bytes', 'KB', 'MB', 'GB', 'TB')
    # base = 1000
    units = ('bytes', 'KiB', 'MiB', 'GiB', 'TiB')
    base = 1024
    if number is None:
        return None
    pow = int(math.log(number, base)) if number > 0 else 0
    pow = min(pow, len(units)-1)
    mantissa = number / (base ** pow)
    return "%.3f %s" % (mantissa, units[pow])


def printUUID(uuid):
    if uuid is None:
        return None
    return "%s...%s" % (uuid[:4], uuid[-4:])


def printVolume(vol):
    """ Return string for dict containing volume info. """
    size = "%s exclusive" % (printBytes(vol['exclusiveSize']))
    if 'totalSize' in vol:
        size = "%s, " % (printBytes(vol['totalSize'])) + size

    return "%s %s (%s)" % (
        vol['path'], vol['uuid'], size,
        )



#! /usr/bin/python

""" Main program to synchronize btrfs snapshots.  See README.md.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
"""

import argparse
import logging
import re
import sys

import BestDiffs
import ButterStore
import S3Store
import Store

theVersion = '0.1'

logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')


def _setupLogging(quietLevel, logFile):
    # theDebugDisplayFormat = '%(levelname)7s:%(filename)s[%(lineno)d] %(funcName)s(): %(message)s'
    theDisplayFormat = '%(message)s'
    theLogFormat = '%(asctime)-15s: %(levelname)7s:%(filename)s[%(lineno)d] %(funcName)s(): %(message)s'

    root = logging.getLogger()
    root.setLevel("DEBUG")

    def add(handler, level, format):
        handler.setLevel(level)
        handler.setFormatter(logging.Formatter(format))
        root.addHandler(handler)

    add(logging.StreamHandler(sys.stdout), "INFO" if quietLevel < 2 else "WARN", theDisplayFormat)

    if logFile is not None:
        add(logging.StreamHandler(logFile), "DEBUG", theLogFormat)

command = argparse.ArgumentParser(
    description="Synchronize two sets of btrfs snapshots.",
    epilog="""
<src>, <dst>:   file:///path/to/directory
                ssh://[user@]host/path/to/directory (Not implemented)
                s3://bucket/prefix[/snapshot]

If only <dst> is supplied, just list available snapshots.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
See README.md and LICENSE.txt for more info.
    """,
    formatter_class=argparse.RawDescriptionHelpFormatter,
)

command.add_argument('source', metavar='<src>', nargs='?',  # nargs='+',
                     help='a source of btrfs snapshots')
command.add_argument('dest', metavar='<dst>',
                     help='the btrfs snapshots to be updated')

command.add_argument('-n', '--dry-run', action="store_true",
                     help="display what would be transferred, but don't do it",
                     )
command.add_argument('-d', '--delete', action="store_true",
                     help='delete any snapshots in <dst> that are not in <src>',
                     )

command.add_argument('-q', '--quiet', action="count", default=0,
                     help="""
                     once: don't display progress.
                     twice: only display error messages""",
                     )
command.add_argument('-l', '--logfile', type=argparse.FileType('w'),
                     help='log debugging information to file',
                     )
command.add_argument('-V', '--version', action="version", version='%(prog)s ' + theVersion,
                     help='display version',
                     )

command.add_argument('--remote-receive', action="store_true",
                     help=argparse.SUPPRESS,
                     )
command.add_argument('--remote-send', action="store_true",
                     help=argparse.SUPPRESS,
                     )
command.add_argument('--remote-list', action="store_true",
                     help=argparse.SUPPRESS,
                     )


def parseSink(uri):
    """ Parse command-line description of sink into a sink object. """
    if uri is None:
        return None

    # logger.debug(uri)
    pattern = re.compile('^(?P<method>[^:/]*)://(?P<host>[^/]*)(/(?P<path>.*))?$')
    match = pattern.match(uri)
    if match is None:
        logger.error("Can't parse snapshot store '%s'", uri)
        return None
    parts = match.groupdict()

    if parts['method'] == 'file':
        parts['path'] = parts['host'] + '/' + parts['path']
    logger.debug(parts)

    Sinks = {
        'file': ButterStore.ButterStore,
        's3': S3Store.S3Store,
        # 'ssh': SSHStore.SSHStore,
    }

    return Sinks[parts['method']](parts['host'], parts['path'])


def main():
    """ Main program. """
    args = command.parse_args()

    _setupLogging(args.quiet, args.logfile)

    logger.debug("Arguments: %s", vars(args))

    progress = args.quiet == 0

    source = parseSink(args.source)

    dest = parseSink(args.dest)

    if source is None:
        for vol in dest.listVolumes():
            print(Store.printVolume(vol))
        return 0

    vols = source.listVolumes()

    best = BestDiffs.BestDiffs([vol['uuid'] for vol in vols], args.delete)
    best.analyze(source, dest)

    summary = best.summary()
    logger.info("Optimal synchronization: %d diffs, %s total",
                summary["count"], Store.humanize(summary["size"]))
    for sink, size in summary["sinks"].items():
        logger.info("%s from %s", Store.humanize(size), sink)

    for diff in best.iterDiffs():
        logger.info("%s: %s", "Keep" if diff.diffSink == dest else "Xfer", diff)

        if diff.diffSink == dest:
            continue

        if args.dry_run:
            continue

        path = diff.diffSink.getVolume(diff.uuid)['path']

        streamContext = dest.receive(diff.uuid, diff.previous, path)

        diff.diffSink.send(diff.uuid, diff.previous, streamContext, progress=progress)

    return 0

if __name__ == "__main__":
    sys.exit(main())



""" Interface to btrfs-tools for snapshots.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
"""

if True:  # Headers
    if True:  # imports
        import subprocess
        import os.path
        import os
        import psutil
        import re
        import sys

    if True:  # constants
        import logging
        logger = logging.getLogger(__name__)
        # logger.setLevel('DEBUG')

        theChunkSize = 100 * (2**20)

        DEVNULL = open(os.devnull, 'wb')


class Butter:

    """ Interface to local btrfs file system snapshots. """

    def __init__(self, path):
        """ Initialize.

        path indicates the btrfs volume, and also the directory containing snapshots.
        """
        self.btrfsVersion = self._getVersion([3, 14])

        if not os.path.isdir(path):
            raise Exception("'%s' is not an existing directory" % (path))

        # userPath - user specified mounted path to directory with snapshots
        self.userPath = path

        # Get tree ID of the containing subvolume of path.
        self.mountID = int(subprocess.check_output(
            ["btrfs", "inspect", "rootid", self.userPath], stderr=sys.stderr
        ).decode("utf-8"))

        # mountPath - mounted path to mount point (doesn't always work!)
        self.mountPath = subprocess.check_output(
            ["df", path], stderr=sys.stderr
        ).decode("utf-8").rsplit(None, 1)[1]

        # relPath - relative path from mountPath to userPath
        self.relPath = os.path.relpath(path, self.mountPath)

        butterPath = subprocess.check_output(
            ["btrfs", "inspect", "subvol", str(self.mountID), self.mountPath], stderr=sys.stderr
        ).decode("utf-8").strip()

        # topPath - btrfs path to immediately enclosing subvolume
        self.topPath = "<FS_TREE>/" + butterPath

        # butterPath - btrfs path to user specified directory with snapshots
        self.butterPath = os.path.normpath(os.path.join(self.topPath, self.relPath))

        logger.debug(
            "MountID: %d, Mount: %s, Path: %s",
            self.mountID, self.mountPath, self.butterPath
            )

        self.volumes = self._getVolumes()

    def listVolumes(self):
        """ List all read-only volumes under path. """
        return self.volumes

    def _getVersion(self, minVersion):
        btrfsVersionString = subprocess.check_output(
            ["btrfs", "--version"], stderr=sys.stderr
            ).decode("utf-8").strip()

        versionPattern = re.compile("[0-9]+(\.[0-9]+)*")
        version = versionPattern.search(btrfsVersionString)

        try:
            version = [int(num) for num in version.group(0).split(".")]
        except AttributeError:
            version = None

        if version < [3, 14]:
            logger.error(
                "%s is not supported.  Please upgrade your btrfs to at least 3.14",
                btrfsVersionString
                )
        else:
            logger.debug("%s", btrfsVersionString)

        return btrfsVersionString

    def _getVolumes(self, readOnly=True):
        vols = {}
        volsByID = {}

        logger.info('Listing "%s" snapshots...', self.relPath)

        subprocess.check_call(["sync"], stderr=sys.stderr)

        result = subprocess.check_output(
            ["btrfs", "sub", "list", "-puta", "-r" if readOnly else "", self.mountPath],
            stderr=sys.stderr
        ).decode("utf-8")

        logger.debug("User path in btrfs: %s", self.butterPath)

        for line in result.splitlines()[2:]:
            logger.debug("%s", line)

            (id, gen, parent, top, uuid, path) = line.split()

            if not path.startswith("<FS_TREE>"):
                if int(top) != self.mountID:
                    logger.error("Can't find absolute path for snapshot %s (%s)", path, id)
                else:
                    path = os.path.normpath(os.path.join(os.path.dirname(self.topPath), path))

            extra = not path.startswith(self.butterPath)

            logger.debug("%s snapshot path in btrfs: %s", "Extra" if extra else "REQUIRED", path)

            if not extra:
                path = os.path.relpath(path, self.butterPath)

            vol = {
                'id': int(id),
                'gen': int(gen),
                'parent': int(parent),
                # 'top': int(top),
                'uuid': uuid,
                'path': path,
                'extra': extra,
            }

            vols[uuid] = vol
            volsByID[int(id)] = vol

        try:
            usage = subprocess.check_output(
                ["btrfs", "qgroup", "show", self.mountPath],
                stderr=sys.stderr)
        except subprocess.CalledProcessError:
            logger.warn("Rescanning subvolume sizes (this may take a while)...")
            subprocess.check_call(
                ["btrfs", "quota", "enable", self.mountPath],
                stderr=sys.stderr)
            subprocess.check_call(
                ["btrfs", "quota", "rescan", "-w", self.mountPath],
                stderr=sys.stderr)
            usage = subprocess.check_output(
                ["btrfs", "qgroup", "show", self.mountPath],
                stderr=sys.stderr)

        for line in usage.splitlines()[2:]:
            (qgroup, totalSize, exclusiveSize) = line.split()
            volID = int(qgroup.split("/")[-1])

            if volID in volsByID:
                logger.debug("Snap info: %s", line)
                volsByID[volID]['totalSize'] = int(totalSize)
                volsByID[volID]['exclusiveSize'] = int(exclusiveSize)

        return vols

    def receive(self):
        """ Return a file-like (stream) object to store a diff. """
        cmd = ["btrfs", "receive", self.userPath]
        subprocess.check_call(["sync"])
        process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stderr=sys.stderr)
        ps = psutil.Process(process.pid)
        ps.ionice(psutil.IOPRIO_CLASS_IDLE)
        return process.stdin

    def _getPath(self, uuid):
        path = self.volumes[uuid]['path']
        if path.startswith('<FS_TREE>'):
            return path[9:]
        else:
            return os.path.normpath(os.path.join(
                self.userPath,
                path
                ))

    def send(self, uuid, parent, streamContext, progress=True):
        """ Write a (incremental) snapshot to the stream context manager. """
        targetPath = self._getPath(uuid)

        if parent is not None:
            cmd = ["btrfs", "send", "-p", self._getPath(parent), targetPath]
        else:
            cmd = ["btrfs", "send", targetPath]

        logger.debug("Command: %s", cmd)

        subprocess.check_call(["sync"], stderr=sys.stderr)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=sys.stderr)
        ps = psutil.Process(process.pid)
        ps.ionice(psutil.IOPRIO_CLASS_IDLE)

        try:
            streamContext.metadata['btrfsVersion'] = self.btrfsVersion
        except AttributeError:
            pass

        try:
            streamContext.progress = progress
        except AttributeError:
            pass

        with streamContext as stream:
            while True:
                data = process.stdout.read(theChunkSize)
                if len(data) == 0:
                    break
                stream.write(data)



""" Back-end for AWS S3.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
"""

# Docs: http://boto.readthedocs.org/en/latest/

from __future__ import division

if True:  # Imports and constants
    if True:  # Imports
        import Store

        import boto
        import datetime
        import io
        import logging
        import os.path
        import pprint
        import re
        import sys
    if True:  # Constants
        # For S3 uploads
        theChunkSize = 100 * 2 ** 20

        # Maximum xumber of progress reports per chunk
        theProgressCount = 50

        # This does transparent S3 server-side encryption
        isEncrypted = True

        logger = logging.getLogger(__name__)
        # logger.setLevel('DEBUG')


class S3Store(Store.Store):

    """ An S3 bucket synchronization source or sink. """

    def __init__(self, host, path):
        """ Initialize.

        host is the bucket name.
        path is an object key prefix to use.
        """
        self.bucketName = host

        if path:
            path = path.strip("/")
        if path:
            path += "/"
        else:
            path = ""
        self.prefix = path

        self.keyPattern = re.compile(S3Store.theKeyPattern % ())

        # List of dict with from, to, size, path
        self.diffs = None

        # { uuid: { path, uuid, } }
        self.vols = None

        logger.info("Listing %s contents...", self)

        try:
            s3 = boto.connect_s3()
        except boto.exception.NoAuthHandlerFound:
            logger.error("Try putting S3 credentials into ~/.boto")
            raise

        self.bucket = s3.get_bucket(self.bucketName)
        self._listBucket()

    def __unicode__(self):
        """ Return text description. """
        return u'S3 Bucket "%s"' % (self.bucketName)

    def __str__(self):
        """ Return text description. """
        return unicode(self).encode('utf-8')

    def _listBucket(self):
        self.vols = {}
        self.diffs = []

        for key in self.bucket.list():
            diff = self._parseKeyName(key.name)

            if diff is None:
                logger.info("Ignoring '%s' in S3", key.name)
                continue

            if diff['from'] == 'None':
                diff['from'] = None

            extra = not diff['fullpath'].startswith(self.prefix.rstrip("/"))

            if not extra:
                diff['path'] = diff['fullpath'][len(self.prefix):]
            else:
                diff['path'] = diff['fullpath']

            if not diff['path']:
                diff['path'] = "."

            self.vols[diff['to']] = {
                'uuid': diff['to'],
                'path': diff['path'],
                'fullpath': diff['fullpath'],
                'extra': extra,
                'exclusiveSize': key.size,
            }

            self.diffs.append({
                'from': diff['from'],
                'to': diff['to'],
                'size': key.size,
                'path': diff['path']
            })

        logger.debug("Diffs:\n%s", pprint.pformat(self.diffs))
        logger.debug("Vols:\n%s", pprint.pformat(self.vols))

    def listVolumes(self):
        """ Return list of volumes that are available. """
        return [vol for vol in self.vols.values() if not vol['extra']]

    def getVolume(self, uuid):
        """ Return info about volume. """
        return self.vols[uuid]

    def iterEdges(self, fromVol):
        """ Return the edges available from fromVol.

        Returned edge is a dict: 'to' UUID, estimated 'size' in bytes
        """
        for diff in self.diffs:
            if diff['from'] == fromVol:
                yield {'to': diff['to'], 'size': diff['size']}

    def hasEdge(self, toUUID, fromUUID):
        """ Test whether edge is in this sink. """
        for diff in self.diffs:
            if diff['from'] == fromUUID and diff['to'] == toUUID:
                return True
        return False

    def receive(self, toUUID, fromUUID, path):
        """ Return a file-like (stream) object to store a diff. """
        path = os.path.normpath(os.path.join(self.prefix, path))
        return _Uploader(self.bucket, self._keyName(toUUID, fromUUID, path))

    theKeyPattern = "^(?P<fullpath>.*)/(?P<to>[-a-zA-Z0-9]*)_(?P<from>[-a-zA-Z0-9]*)$"

    def _keyName(self, toUUID, fromUUID, path):
        return "%s/%s_%s" % (path, toUUID, fromUUID)

    def _parseKeyName(self, name):
        match = self.keyPattern.match(name)
        return match.groupdict() if match else None

    def send(self, toUUID, fromUUID, streamContext, progress=True):
        """ Write the diff to the streamContext. """
        path = self.vols[toUUID]['fullpath']
        key = self.bucket.get_key(self._keyName(toUUID, fromUUID, path))

        with streamContext as stream:
            if progress:
                key.get_contents_to_file(stream, cb=_DisplayProgress(), num_cb=theProgressCount)
            else:
                key.get_contents_to_file(stream)

        if sys.stdout.isatty():
            sys.stdout.write("\n")

    def _upload(self, stream, keyName):
        # key = self.bucket.get_key(keyName)
        # key = self.bucket.new_key(keyName)

        # set_contents_from_stream is not supported for S3
        # key.set_contents_from_stream(stream, replace=False, cb=displayProgress, size=1000)
        # key.set_contents_from_filename(fileName, replace=False, cb=displayProgress)

        with _Uploader(self.bucket, keyName) as uploader:
            while True:
                data = stream.read(theChunkSize)
                if not data:
                    break
                uploader.upload(data)


class _DisplayProgress:
    def __init__(self):
        self.startTime = datetime.datetime.now()

    def __call__(self, sent, total):
        if not sys.stdout.isatty():
            return

        elapsed = datetime.datetime.now() - self.startTime
        mbps = (sent * 8 / (10**6) / elapsed.total_seconds())

        sys.stdout.write(
            "\r %s: Sent %s of %s (%d%%) (%.3g Mbps) %20s\r" % (
                elapsed,
                Store.humanize(sent),
                Store.humanize(total),
                int(100 * sent / total),
                mbps,
                " ",
                )
        )

        sys.stdout.flush()


class _Uploader:

    def __init__(self, bucket, keyName):
        self.bucket = bucket
        self.keyName = keyName
        self.uploader = None
        self.chunkCount = None
        self.metadata = {}
        self.progress = True

    def __enter__(self):
        self.open()
        return self

    def open(self):
        self.uploader = self.bucket.initiate_multipart_upload(
            self.keyName,
            encrypt_key=isEncrypted,
            metadata=self.metadata,
        )
        self.chunkCount = 0

    def __exit__(self, exceptionType, exceptionValue, traceback):
        self.close(abort=exceptionType is not None)
        return False  # Don't supress exception

    def write(self, bytes):
        self.upload(bytes)

    def close(self, abort=False):
        if not abort:
            self.uploader.complete_upload()
            # You cannot change metadata after uploading
            # if self.metadata:
            #     key = self.bucket.get_key(self.keyName)
            #     key.update_metadata(self.metadata)
        else:
            # TODO: this doesn't free storage used by part uploads currently in progress
            self.uploader.cancel_upload()
        self.uploader = None

    def fileno(self):
        raise IOError("S3 uploads don't use file numbers.")

    def upload(self, bytes):
        self.chunkCount += 1
        logger.info(
            "Uploading %s chunk #%d for %s",
            Store.humanize(len(bytes)), self.chunkCount, self.keyName
        )
        fileObject = io.BytesIO(bytes)

        if self.progress:
            self.uploader.upload_part_from_file(
                fileObject, self.chunkCount, cb=_DisplayProgress(), num_cb=theProgressCount
            )
            if sys.stdout.isatty():
                sys.stdout.write("\n")
        else:
            self.uploader.upload_part_from_file(
                fileObject, self.chunkCount
            )



""" Manage read-only volumes in local btrfs file system.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
"""

from __future__ import division

import Butter
import Store

import math
import logging
logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')

theMinimumChangeRate = .00001


class ButterStore(Store.Store):

    """ A local btrfs synchronization source or sink. """

    def __init__(self, host, path):
        """ Initialize.

        host is ignored.
        path is the file system location of the read-only subvolumes.
        """
        self.path = path
        self.butter = Butter.Butter(path)
        self.volumes = self.butter.listVolumes()

    def __unicode__(self):
        """ English description of self. """
        return u"btrfs %s" % (self.path)

    def __str__(self):
        """ English description of self. """
        return unicode(self).encode('utf-8')

    def listVolumes(self):
        """ Return list of volumes that are available. """
        return [vol for vol in self.volumes.values() if not vol['extra']]

    def getVolume(self, uuid):
        """ Return dict of info for a specific volume. """
        return self.volumes[uuid]

    def iterEdges(self, fromVol):
        """ Return the edges available from fromVol.

        Returned edge is a dict: 'to' UUID, estimated 'size' in bytes
        """
        if fromVol is None:
            for toVol in self.volumes.values():
                yield {'to': toVol['uuid'], 'size': toVol['totalSize']}
            return

        if fromVol not in self.volumes:
            return

        fromVol = self.volumes[fromVol]
        fromParent = fromVol['parent']

        vols = [toVol for toVol in self.volumes.values()
                if toVol['parent'] == fromParent
                ]

        changeRate = self._calcChangeRate(vols)

        for toVol in vols:
            if toVol == fromVol:
                continue

            # This gives a conservative estimate of the size of the diff

            estimatedSize = self._estimateSize(toVol, fromVol, changeRate)

            yield {'to': toVol['uuid'], 'size': estimatedSize}

    def hasEdge(self, toUUID, fromUUID):
        """ Store already contains this edge. """
        return toUUID in self.volumes and fromUUID in self.volumes

    def receive(self, toUUID, fromUUID, path):
        """ Return a file-like (stream) object to store a diff. """
        return self.butter.receive()

    def _estimateSize(self, toVol, fromVol, changeRate):
        fromGen = fromVol['gen']
        genDiff = abs(toVol['gen'] - fromGen)

        estimatedSize = max(0, toVol['totalSize'] - fromVol['totalSize'])
        estimatedSize += toVol['totalSize'] * (1 - math.exp(-changeRate * genDiff))
        estimatedSize = max(toVol['exclusiveSize'], estimatedSize)

        return estimatedSize

    def _calcChangeRate(self, vols):
        total = 0
        diffs = 0
        minGen = vols[0]['gen']
        maxGen = minGen
        minSize = vols[0]['totalSize']
        maxSize = minSize

        for vol in vols:
            total += vol['totalSize']
            diffs += vol['exclusiveSize']
            minGen = min(minGen, vol['gen'])
            maxGen = max(maxGen, vol['gen'])
            minSize = min(minSize, vol['totalSize'])
            maxSize = max(maxSize, vol['totalSize'])

        try:
            # exclusiveSize is often useless,
            # because data may be shared with read-write volumes not usable for send operations
            diffs = max(diffs, maxSize - minSize)
            rate = - math.log(1 - diffs / total) * (len(vols) - 1) / (maxGen - minGen)
            rate /= 10  # Fudge
        except (ZeroDivisionError, ValueError):
            logger.debug("Using minimum change rate.")
            rate = theMinimumChangeRate

        logger.debug("Change rate: %f", rate)

        return rate

    def send(self, toUUID, fromUUID, streamContext, progress=True):
        """ Write the diff to the stream. """
        self.butter.send(toUUID, fromUUID, streamContext, progress)



#! /usr/bin/python

""" Package setup for buttersink. """

from setuptools import setup


setup(
    name="buttersink",
    version="0.1",
    packages=['buttersink'],

    # metadata for upload to PyPI
    author="Ames Cornish",
    author_email="buttersink@montebellopartners.com",
    description="Buttersink is like rsync for btrfs snapshots",
    license="GPLv3",
    keywords="btrfs sync synchronize rsync snapshot subvolume buttersink",
    url="https://github.com/AmesCornish/buttersink/wiki",
    # could also include long_description, download_url, classifiers, etc.

    entry_points={
        'console_scripts': ['buttersink=buttersink.buttersink:main'],
    },

    requires=['boto', 'dev', 'psutil'],
    
    # scripts=['buttersink.py'],
    # package_dir={'buttersink': '..'},
    # include_package_data=True,

    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    # install_requires=['docutils>=0.3'],

)
